"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ethUtil = require("ethereumjs-util");
const Web3 = require("web3");
const bitstream_1 = require("./bitstream");
const logs_1 = require("./logs");
const types_1 = require("./types");
class MultiHashUtil {
    constructor() {
        try {
            if (web3) {
                this.web3Instance = web3;
            }
            else {
                this.web3Instance = new Web3(new Web3.providers.HttpProvider("http://localhost:8545"));
            }
        }
        catch (err) {
            logs_1.logDebug("get web3 instance in Order class failed. err:", err);
        }
    }
    signOrderAsync(order) {
        return __awaiter(this, void 0, void 0, function* () {
            const signer = order.owner;
            return yield this.signAsync(order.signAlgorithm, order.hash, signer, order.signerPrivateKey);
        });
    }
    signAsync(algorithm, hash, address, privateKey) {
        return __awaiter(this, void 0, void 0, function* () {
            // Default to standard Ethereum signing
            algorithm = Object.is(algorithm, undefined) ? types_1.SignAlgorithm.Ethereum : algorithm;
            const sig = new bitstream_1.Bitstream();
            sig.addNumber(algorithm, 1);
            switch (+algorithm) {
                case types_1.SignAlgorithm.Ethereum:
                    yield this.signEthereumAsync(sig, hash, privateKey);
                    return sig.getData();
                case types_1.SignAlgorithm.EIP712:
                    yield this.signEIP712Async(sig, hash, privateKey);
                    return sig.getData();
                case types_1.SignAlgorithm.None:
                    return null;
                default:
                    throw Error("Unsupported hashing algorithm: " + algorithm);
            }
        });
    }
    verifySignature(signer, hash, multihash) {
        const bitstream = new bitstream_1.Bitstream(multihash);
        assert(bitstream.length() >= 2, "invalid multihash format");
        const algorithm = bitstream.extractUint8(0);
        const size = bitstream.extractUint8(1);
        assert.equal(bitstream.length(), (2 + size), "bad multihash size");
        if (algorithm === types_1.SignAlgorithm.Ethereum) {
            assert.notEqual(signer, "0x0", "invalid signer address");
            assert.equal(size, 65, "bad Ethereum multihash size");
            const v = bitstream.extractUint8(2);
            const r = bitstream.extractBytes32(3);
            const s = bitstream.extractBytes32(3 + 32);
            try {
                const msgHash = ethUtil.hashPersonalMessage(hash);
                const pub = ethUtil.ecrecover(msgHash, v, r, s);
                const recoveredAddress = "0x" + ethUtil.pubToAddress(pub).toString("hex");
                return signer.toLowerCase() === recoveredAddress.toLowerCase();
            }
            catch (_a) {
                return false;
            }
        }
        else if (algorithm === types_1.SignAlgorithm.EIP712) {
            assert.notEqual(signer, "0x0", "invalid signer address");
            assert.equal(size, 65, "bad EIP712 multihash size");
            const v = bitstream.extractUint8(2);
            const r = bitstream.extractBytes32(3);
            const s = bitstream.extractBytes32(3 + 32);
            try {
                const pub = ethUtil.ecrecover(hash, v, r, s);
                const recoveredAddress = "0x" + ethUtil.pubToAddress(pub).toString("hex");
                return signer.toLowerCase() === recoveredAddress.toLowerCase();
            }
            catch (_b) {
                return false;
            }
        }
        else {
            return false;
        }
    }
    signEthereumAsync(sig, hash, privateKey) {
        return __awaiter(this, void 0, void 0, function* () {
            const parts = [Buffer.from("\x19Ethereum Signed Message:\n32", "utf8"), hash];
            const totalHash = ethUtil.sha3(Buffer.concat(parts));
            const signature = ethUtil.ecsign(totalHash, new Buffer(privateKey, "hex"));
            sig.addNumber(1 + 32 + 32, 1);
            sig.addNumber(signature.v, 1);
            sig.addHex(ethUtil.bufferToHex(signature.r));
            sig.addHex(ethUtil.bufferToHex(signature.s));
        });
    }
    signEIP712Async(sig, hash, privateKey) {
        return __awaiter(this, void 0, void 0, function* () {
            const signature = ethUtil.ecsign(hash, new Buffer(privateKey, "hex"));
            sig.addNumber(1 + 32 + 32, 1);
            sig.addNumber(signature.v, 1);
            sig.addHex(ethUtil.bufferToHex(signature.r));
            sig.addHex(ethUtil.bufferToHex(signature.s));
        });
    }
}
exports.MultiHashUtil = MultiHashUtil;
//# sourceMappingURL=multihash.js.map