"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignumber_js_1 = require("bignumber.js");
const BN = require("bn.js");
const bitstream_1 = require("./bitstream");
const eip712_1 = require("./eip712");
const ensure_1 = require("./ensure");
const multihash_1 = require("./multihash");
const types_1 = require("./types");
class OrderUtil {
    constructor(context) {
        this.zeroBytes32 = "0x" + "0".repeat(64);
        this.context = context;
    }
    updateBrokerAndInterceptor(order) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!order.broker) {
                order.broker = order.owner;
            }
            else {
                const returnValue = yield this.context.orderBrokerRegistry.methods.getBroker(order.owner, order.broker).call();
                order.valid = order.valid && ensure_1.ensure(returnValue.registered, "order broker is not registered");
                // order.brokerInterceptor =
                //   (brokerInterceptor === "0x0000000000000000000000000000000000000000") ?
                //   undefined : returnValue.brokerInterceptor;
            }
        });
    }
    validateInfo(order) {
        return __awaiter(this, void 0, void 0, function* () {
            let valid = true;
            valid = valid && ensure_1.ensure(order.version === 0, "unsupported order version");
            valid = valid && ensure_1.ensure(order.owner ? true : false, "invalid order owner");
            valid = valid && ensure_1.ensure(order.tokenS ? true : false, "invalid order tokenS");
            valid = valid && ensure_1.ensure(order.tokenB ? true : false, "invalid order tokenB");
            valid = valid && ensure_1.ensure(order.amountS !== 0, "invalid order amountS");
            valid = valid && ensure_1.ensure(order.amountB !== 0, "invalid order amountB");
            valid = valid && ensure_1.ensure(order.waiveFeePercentage <= this.context.feePercentageBase, "invalid waive percentage");
            valid = valid && ensure_1.ensure(order.waiveFeePercentage >= -this.context.feePercentageBase, "invalid waive percentage");
            valid = valid && ensure_1.ensure(order.tokenSFeePercentage < this.context.feePercentageBase, "invalid tokenS percentage");
            valid = valid && ensure_1.ensure(order.tokenBFeePercentage < this.context.feePercentageBase, "invalid tokenB percentage");
            valid = valid && ensure_1.ensure(order.walletSplitPercentage <= 100, "invalid wallet split percentage");
            if (order.dualAuthAddr) {
                valid = valid && ensure_1.ensure(order.dualAuthSig && order.dualAuthSig.length > 0, "missing dual author signature");
            }
            const blockTimestamp = this.context.blockTimestamp;
            valid = valid && ensure_1.ensure(order.validSince <= blockTimestamp, "order is too early to match");
            valid = valid && ensure_1.ensure(order.validUntil ? order.validUntil > blockTimestamp : true, "order is expired");
            // We only support ERC20 for now
            // valid = valid && ensure(
            //   (order.tokenTypeS === TokenType.ERC20 && order.trancheS === this.zeroBytes32),
            //   "invalid trancheS",
            // );
            // valid = valid && ensure(
            //   (order.tokenTypeB === TokenType.ERC20 && order.trancheB === this.zeroBytes32),
            //   "invalid trancheB",
            // );
            valid = valid && ensure_1.ensure((order.tokenTypeFee === types_1.TokenType.ERC20), "invalid tokenTypeFee");
            // valid = valid && ensure(
            //   (order.tokenTypeS === TokenType.ERC20 && order.transferDataS === "0x"),
            //   "invalid transferDataS",
            // );
            // This emulates the revert in solidity with invalid enum values
            assert(order.tokenTypeS < types_1.TokenType.COUNT, "invalid opcode");
            assert(order.tokenTypeB < types_1.TokenType.COUNT, "invalid opcode");
            assert(order.tokenTypeFee < types_1.TokenType.COUNT, "invalid opcode");
            order.valid = order.valid && valid;
        });
    }
    validateAllOrNone(order) {
        order.valid = order.valid && ensure_1.ensure(order.filledAmountS.eq(order.amountS), "allOrNone not completely filled");
    }
    checkBrokerSignature(order) {
        return __awaiter(this, void 0, void 0, function* () {
            let signatureValid = true;
            // If the order was already partially filled we don't have to check the signature again
            if (order.filledAmountS.gt(0)) {
                signatureValid = true;
            }
            else if (!order.sig) {
                const orderHashHex = "0x" + order.hash.toString("hex");
                const isRegistered = yield this.context.orderRegistry.methods.isOrderHashRegistered(order.broker, orderHashHex).call();
                const isOnchainOrder = yield this.context.orderBook.methods.orderSubmitted(orderHashHex).call();
                signatureValid = isRegistered || isOnchainOrder;
            }
            else {
                const multiHashUtil = new multihash_1.MultiHashUtil();
                signatureValid = multiHashUtil.verifySignature(order.broker, order.hash, order.sig);
            }
            order.valid = order.valid && ensure_1.ensure(signatureValid, "invalid order signature");
        });
    }
    checkDualAuthSignature(order, miningHash) {
        if (order.dualAuthSig) {
            const multiHashUtil = new multihash_1.MultiHashUtil();
            const signatureValid = multiHashUtil.verifySignature(order.dualAuthAddr, miningHash, order.dualAuthSig);
            order.valid = order.valid && ensure_1.ensure(signatureValid, "invalid order dual auth signature");
        }
    }
    toTypedData(order) {
        const typedData = {
            types: {
                EIP712Domain: [
                    { name: "name", type: "string" },
                    { name: "version", type: "string" },
                ],
                Order: [
                    { name: "amountS", type: "uint" },
                    { name: "amountB", type: "uint" },
                    { name: "feeAmount", type: "uint" },
                    { name: "validSince", type: "uint" },
                    { name: "validUntil", type: "uint" },
                    { name: "owner", type: "address" },
                    { name: "tokenS", type: "address" },
                    { name: "tokenB", type: "address" },
                    { name: "dualAuthAddr", type: "address" },
                    { name: "broker", type: "address" },
                    { name: "orderInterceptor", type: "address" },
                    { name: "wallet", type: "address" },
                    { name: "tokenRecipient", type: "address" },
                    { name: "feeToken", type: "address" },
                    { name: "walletSplitPercentage", type: "uint16" },
                    { name: "tokenSFeePercentage", type: "uint16" },
                    { name: "tokenBFeePercentage", type: "uint16" },
                    { name: "allOrNone", type: "bool" },
                    { name: "tokenTypeS", type: "uint8" },
                    { name: "tokenTypeB", type: "uint8" },
                    { name: "tokenTypeFee", type: "uint8" },
                    { name: "trancheS", type: "bytes32" },
                    { name: "trancheB", type: "bytes32" },
                    { name: "transferDataS", type: "bytes" },
                ],
            },
            primaryType: "Order",
            domain: {
                name: "Loopring Protocol",
                version: "2",
            },
            message: {
                amountS: this.toBN(order.amountS),
                amountB: this.toBN(order.amountB),
                feeAmount: this.toBN(order.feeAmount),
                validSince: order.validSince ? this.toBN(order.validSince) : this.toBN(0),
                validUntil: order.validUntil ? this.toBN(order.validUntil) : this.toBN(0),
                owner: order.owner,
                tokenS: order.tokenS,
                tokenB: order.tokenB,
                dualAuthAddr: order.dualAuthAddr ? order.dualAuthAddr : "",
                broker: order.broker ? order.broker : "",
                orderInterceptor: order.orderInterceptor ? order.orderInterceptor : "",
                wallet: order.walletAddr ? order.walletAddr : "",
                tokenRecipient: order.tokenRecipient,
                feeToken: order.feeToken,
                walletSplitPercentage: order.walletSplitPercentage,
                tokenSFeePercentage: order.tokenSFeePercentage,
                tokenBFeePercentage: order.tokenBFeePercentage,
                allOrNone: order.allOrNone,
                tokenTypeS: order.tokenTypeS ? order.tokenTypeS : types_1.TokenType.ERC20,
                tokenTypeB: order.tokenTypeB ? order.tokenTypeB : types_1.TokenType.ERC20,
                tokenTypeFee: order.tokenTypeFee ? order.tokenTypeFee : types_1.TokenType.ERC20,
                trancheS: order.trancheS ? order.trancheS : "",
                trancheB: order.trancheB ? order.trancheB : "",
                transferDataS: order.transferDataS ? order.transferDataS : "",
            },
        };
        return typedData;
    }
    toTypedDataJSON(order) {
        // BN outputs hex numbers in toJSON, but signTypedData expects decimal numbers
        const replacer = (key, value) => {
            if (key === "amountS" || key === "amountB" || key === "feeAmount" ||
                key === "validSince" || key === "validUntil") {
                return "" + parseInt(value, 16);
            }
            return value;
        };
        const typedData = this.toTypedData(order);
        const json = JSON.stringify(typedData, replacer);
        return json;
    }
    getOrderHash(order) {
        const typedData = this.toTypedData(order);
        const orderHash = eip712_1.getEIP712Message(typedData);
        return orderHash;
    }
    toOrderBookSubmitParams(orderInfo) {
        const emptyAddr = "0x" + "0".repeat(40);
        const zeroBytes32 = "0x" + "0".repeat(64);
        const data = new bitstream_1.Bitstream();
        data.addAddress(orderInfo.owner, 32);
        data.addAddress(orderInfo.tokenS, 32);
        data.addAddress(orderInfo.tokenB, 32);
        data.addNumber(orderInfo.amountS, 32);
        data.addNumber(orderInfo.amountB, 32);
        data.addNumber(orderInfo.validSince, 32);
        data.addAddress(orderInfo.broker ? orderInfo.broker : emptyAddr, 32);
        data.addAddress(orderInfo.orderInterceptor ? orderInfo.orderInterceptor : emptyAddr, 32);
        data.addAddress(orderInfo.walletAddr ? orderInfo.walletAddr : emptyAddr, 32);
        data.addNumber(orderInfo.validUntil ? orderInfo.validUntil : 0, 32);
        data.addNumber(orderInfo.allOrNone ? 1 : 0, 32);
        data.addAddress(orderInfo.feeToken ? orderInfo.feeToken : emptyAddr, 32);
        data.addNumber(orderInfo.feeAmount ? orderInfo.feeAmount : 0, 32);
        data.addNumber(orderInfo.tokenSFeePercentage ? orderInfo.tokenSFeePercentage : 0, 32);
        data.addNumber(orderInfo.tokenBFeePercentage ? orderInfo.tokenBFeePercentage : 0, 32);
        data.addAddress(orderInfo.tokenRecipient ? orderInfo.tokenRecipient : orderInfo.owner, 32);
        data.addNumber(orderInfo.walletSplitPercentage ? orderInfo.walletSplitPercentage : 0, 32);
        data.addNumber(orderInfo.tokenTypeS ? orderInfo.tokenTypeS : types_1.TokenType.ERC20, 32);
        data.addNumber(orderInfo.tokenTypeB ? orderInfo.tokenTypeB : types_1.TokenType.ERC20, 32);
        data.addNumber(orderInfo.tokenTypeFee ? orderInfo.tokenTypeFee : types_1.TokenType.ERC20, 32);
        data.addHex(orderInfo.trancheS ? orderInfo.trancheS : zeroBytes32);
        data.addHex(orderInfo.trancheB ? orderInfo.trancheB : zeroBytes32);
        if (orderInfo.transferDataS) {
            data.addNumber((orderInfo.transferDataS.length - 2) / 2, 32);
            data.addHex(orderInfo.transferDataS);
        }
        else {
            data.addNumber(0, 32);
        }
        return data.getData();
    }
    checkP2P(orderInfo) {
        orderInfo.P2P = (orderInfo.tokenSFeePercentage > 0 || orderInfo.tokenBFeePercentage > 0);
    }
    getSpendableS(order) {
        return __awaiter(this, void 0, void 0, function* () {
            const spendable = yield this.getSpendable(order.tokenS, order.owner, order.broker, order.brokerInterceptor, order.tokenSpendableS, order.brokerSpendableS);
            return spendable;
        });
    }
    getSpendableFee(order) {
        return __awaiter(this, void 0, void 0, function* () {
            const spendable = yield this.getSpendable(order.feeToken, order.owner, order.broker, order.brokerInterceptor, order.tokenSpendableFee, order.brokerSpendableFee);
            return spendable;
        });
    }
    reserveAmountS(order, amount) {
        return __awaiter(this, void 0, void 0, function* () {
            const spendableS = yield this.getSpendableS(order);
            assert(spendableS.gte(amount), "spendableS >= reserve amount");
            order.tokenSpendableS.reserved = order.tokenSpendableS.reserved.plus(amount);
            if (order.brokerInterceptor) {
                order.brokerSpendableS.reserved = order.brokerSpendableS.reserved.plus(amount);
            }
        });
    }
    reserveAmountFee(order, amount) {
        return __awaiter(this, void 0, void 0, function* () {
            assert((yield this.getSpendableFee(order)).gte(amount), "spendableFee >= reserve amount");
            order.tokenSpendableFee.reserved = order.tokenSpendableFee.reserved.plus(amount);
            if (order.brokerInterceptor) {
                order.brokerSpendableFee.reserved = order.brokerSpendableFee.reserved.plus(amount);
            }
        });
    }
    resetReservations(order) {
        order.tokenSpendableS.reserved = new bignumber_js_1.BigNumber(0);
        order.tokenSpendableFee.reserved = new bignumber_js_1.BigNumber(0);
        if (order.brokerInterceptor) {
            order.tokenSpendableS.reserved = new bignumber_js_1.BigNumber(0);
            order.tokenSpendableFee.reserved = new bignumber_js_1.BigNumber(0);
        }
    }
    getERC20Spendable(spender, tokenAddress, owner) {
        return __awaiter(this, void 0, void 0, function* () {
            const token = this.context.ERC20Contract;
            token.options.address = tokenAddress;
            const balance = yield token.methods.balanceOf(owner).call();
            const allowance = yield token.methods.allowance(owner, spender).call();
            const spendable = bignumber_js_1.BigNumber.min(balance, allowance);
            return new bignumber_js_1.BigNumber(spendable.toString());
        });
    }
    getBrokerAllowance(tokenAddr, owner, broker, brokerInterceptor) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this.context.BrokerInterceptorContract.at(brokerInterceptor).getAllowance(owner, broker, tokenAddr);
            }
            catch (_a) {
                return new bignumber_js_1.BigNumber(0);
            }
        });
    }
    getSpendable(token, owner, broker, brokerInterceptor, tokenSpendable, brokerSpendable) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!tokenSpendable.initialized) {
                tokenSpendable.amount = yield this.getERC20Spendable(this.context.tradeDelegate.options.address, token, owner);
                tokenSpendable.initialized = true;
                // Testing
                tokenSpendable.initialAmount = tokenSpendable.amount;
            }
            let spendable = tokenSpendable.amount.minus(tokenSpendable.reserved);
            assert(spendable.gte(0), "spendable >= 0");
            if (brokerInterceptor) {
                if (!brokerSpendable.initialized) {
                    brokerSpendable.amount = yield this.getBrokerAllowance(token, owner, broker, brokerInterceptor);
                    brokerSpendable.initialized = true;
                    // Testing
                    brokerSpendable.initialAmount = brokerSpendable.amount;
                }
                const brokerSpendableAmount = brokerSpendable.amount.minus(brokerSpendable.reserved);
                assert(brokerSpendableAmount.gte(0), "brokerSpendable >= 0");
                spendable = (brokerSpendableAmount.lt(spendable)) ? brokerSpendableAmount : spendable;
            }
            return spendable;
        });
    }
    toBN(n) {
        return new BN((new bignumber_js_1.BigNumber(n)).toString(10), 10);
    }
}
exports.OrderUtil = OrderUtil;
//# sourceMappingURL=order.js.map