"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignumber_js_1 = require("bignumber.js");
const BN = require("bn.js");
const bitstream_1 = require("./bitstream");
const ensure_1 = require("./ensure");
const exchange_deserializer_1 = require("./exchange_deserializer");
const mining_1 = require("./mining");
const order_1 = require("./order");
const ring_1 = require("./ring");
const xor_1 = require("./xor");
class ProtocolSimulator {
    constructor(context) {
        this.offLineMode = false;
        this.context = context;
        this.orderUtil = new order_1.OrderUtil(context);
    }
    deserialize(data, transactionOrigin) {
        const exchangeDeserializer = new exchange_deserializer_1.ExchangeDeserializer(this.context);
        const [mining, orders, rings] = exchangeDeserializer.deserialize(data);
        const ringsInfo = {
            rings,
            orders,
            feeRecipient: mining.feeRecipient,
            miner: mining.miner,
            sig: mining.sig,
            transactionOrigin,
        };
        return ringsInfo;
    }
    simulateAndReport(ringsInfo) {
        return __awaiter(this, void 0, void 0, function* () {
            const mining = new mining_1.Mining(this.context, ringsInfo.feeRecipient ? ringsInfo.feeRecipient : ringsInfo.transactionOrigin, ringsInfo.miner, ringsInfo.sig);
            const orders = ringsInfo.orders;
            const rings = [];
            for (const indexes of ringsInfo.rings) {
                const ringOrders = [];
                for (const orderIndex of indexes) {
                    const orderInfo = ringsInfo.orders[orderIndex];
                    ringOrders.push(orderInfo);
                }
                const ring = new ring_1.Ring(this.context, ringOrders);
                rings.push(ring);
            }
            for (const order of orders) {
                order.valid = true;
                yield this.orderUtil.validateInfo(order);
                this.orderUtil.checkP2P(order);
                order.hash = this.orderUtil.getOrderHash(order);
                yield this.orderUtil.updateBrokerAndInterceptor(order);
            }
            yield this.batchGetFilledAndCheckCancelled(orders);
            this.updateBrokerSpendables(orders);
            for (const order of orders) {
                yield this.orderUtil.checkBrokerSignature(order);
            }
            for (let i = 0; i < orders.length; i++) {
                // An order can only be sent once
                for (let j = i + 1; j < orders.length; j++) {
                    assert(orders[i].hash !== orders[j].hash, "INVALID_VALUE");
                }
            }
            for (const ring of rings) {
                ring.updateHash();
            }
            mining.updateHash(rings);
            yield mining.updateMinerAndInterceptor();
            assert(mining.checkMinerSignature(ringsInfo.transactionOrigin) === true, "INVALID_SIG");
            for (const order of orders) {
                this.orderUtil.checkDualAuthSignature(order, mining.hash);
            }
            const ringMinedEvents = [];
            const invalidRingEvents = [];
            const transferItems = [];
            const feeBalances = {};
            for (const ring of rings) {
                ring.checkOrdersValid();
                ring.checkForSubRings();
                yield ring.calculateFillAmountAndFee();
                if (ring.valid) {
                    ring.adjustOrderStates();
                }
            }
            // Check if the allOrNone orders are completely filled over all rings
            // This can invalidate rings
            this.checkRings(orders, rings);
            for (const ring of rings) {
                if (ring.valid) {
                    const ringReport = yield this.simulateAndReportSingle(ring, mining, feeBalances);
                    ringMinedEvents.push(ringReport.ringMinedEvent);
                    // Merge transfer items if possible
                    for (const ringTransferItem of ringReport.transferItems) {
                        let addNew = true;
                        for (const transferItem of transferItems) {
                            if (transferItem.token === ringTransferItem.token &&
                                transferItem.from === ringTransferItem.from &&
                                transferItem.to === ringTransferItem.to) {
                                transferItem.amount = transferItem.amount.plus(ringTransferItem.amount);
                                addNew = false;
                            }
                        }
                        if (addNew) {
                            transferItems.push(ringTransferItem);
                        }
                    }
                }
                else {
                    const invalidRingEvent = {
                        ringHash: "0x" + ring.hash.toString("hex"),
                    };
                    invalidRingEvents.push(invalidRingEvent);
                }
            }
            const report = yield this.collectReport(ringsInfo, mining, rings, transferItems, feeBalances, ringMinedEvents, invalidRingEvents);
            yield this.validateRings(ringsInfo, report);
            return report;
        });
    }
    checkRings(orders, rings) {
        return __awaiter(this, void 0, void 0, function* () {
            // Check if allOrNone orders are completely filled
            // When a ring is turned invalid because of an allOrNone order we have to
            // recheck the other rings again because they may contain other allOrNone orders
            // that may not be completely filled anymore.
            let reevaluateRings = true;
            while (reevaluateRings) {
                reevaluateRings = false;
                for (const order of orders) {
                    // Check if this order needs to be completely filled
                    if (order.allOrNone) {
                        const validBefore = order.valid;
                        this.orderUtil.validateAllOrNone(order);
                        // Check if the order valid status has changed
                        reevaluateRings = reevaluateRings || (validBefore !== order.valid);
                    }
                }
                if (reevaluateRings) {
                    for (const ring of rings) {
                        const validBefore = ring.valid;
                        ring.checkOrdersValid();
                        // If the ring was valid before the completely filled check we have to revert the filled amountS
                        // of the orders in the ring. This is a bit awkward so maybe there's a better solution.
                        if (!ring.valid && validBefore) {
                            ring.revertOrderStats();
                        }
                    }
                }
            }
        });
    }
    simulateAndReportSingle(ring, mining, feeBalances) {
        return __awaiter(this, void 0, void 0, function* () {
            const transferItems = yield ring.doPayments(mining, feeBalances);
            const fills = ring.generateFills();
            const ringMinedEvent = {
                ringIndex: new bignumber_js_1.BigNumber(this.context.ringIndex++),
                ringHash: "0x" + ring.hash.toString("hex"),
                feeRecipient: mining.feeRecipient,
                fills,
            };
            return { ringMinedEvent, transferItems };
        });
    }
    batchGetFilledAndCheckCancelled(orders) {
        return __awaiter(this, void 0, void 0, function* () {
            const bitstream = new bitstream_1.Bitstream();
            for (const order of orders) {
                bitstream.addAddress(order.broker, 32);
                bitstream.addAddress(order.owner, 32);
                bitstream.addHex(order.hash.toString("hex"));
                bitstream.addNumber(order.validSince, 32);
                bitstream.addHex(xor_1.xor(order.tokenS, order.tokenB, 20));
                bitstream.addNumber(0, 12);
            }
            const fills = yield this.context.tradeHistory.methods.batchGetFilledAndCheckCancelled(bitstream.getBytes32Array()).call();
            const cancelledValue = new BN("F".repeat(64), 16);
            for (const [i, order] of orders.entries()) {
                const fillBN = new BN(new bignumber_js_1.BigNumber(fills[i].toString(16)).toString(16), 16);
                order.filledAmountS = new bignumber_js_1.BigNumber(fillBN.toString());
                order.valid = order.valid && ensure_1.ensure(!fillBN.eq(cancelledValue), "order is cancelled");
            }
        });
    }
    updateBrokerSpendables(orders) {
        const brokerSpendables = [];
        const addBrokerSpendable = (broker, owner, token) => {
            // Find an existing one
            for (const spendable of brokerSpendables) {
                if (spendable.broker === broker && spendable.owner === owner && spendable.token === token) {
                    return spendable.spendable;
                }
            }
            // Create a new one
            const newSpendable = {
                initialized: false,
                amount: new bignumber_js_1.BigNumber(0),
                reserved: new bignumber_js_1.BigNumber(0),
            };
            const newBrokerSpendable = {
                broker,
                owner,
                token,
                spendable: newSpendable,
            };
            brokerSpendables.push(newBrokerSpendable);
            return newBrokerSpendable.spendable;
        };
        for (const order of orders) {
            if (order.brokerInterceptor) {
                order.brokerSpendableS = addBrokerSpendable(order.broker, order.owner, order.tokenS);
                order.brokerSpendableFee = addBrokerSpendable(order.broker, order.owner, order.feeToken);
            }
        }
    }
    collectReport(ringsInfo, mining, rings, transferItems, feeBalances, ringMinedEvents, invalidRingEvents) {
        return __awaiter(this, void 0, void 0, function* () {
            const orders = ringsInfo.orders;
            // Collect balances before the transaction
            const balancesBefore = {};
            for (const order of orders) {
                if (!balancesBefore[order.tokenS]) {
                    balancesBefore[order.tokenS] = {};
                }
                if (!balancesBefore[order.tokenB]) {
                    balancesBefore[order.tokenB] = {};
                }
                if (!balancesBefore[order.feeToken]) {
                    balancesBefore[order.feeToken] = {};
                }
                if (!balancesBefore[order.tokenS][order.owner]) {
                    balancesBefore[order.tokenS][order.owner] =
                        yield this.orderUtil.getERC20Spendable(this.context.tradeDelegate.options.address, order.tokenS, order.owner);
                }
                if (!balancesBefore[order.tokenB][order.tokenRecipient]) {
                    balancesBefore[order.tokenB][order.tokenRecipient] =
                        yield this.orderUtil.getERC20Spendable(this.context.tradeDelegate.options.address, order.tokenB, order.tokenRecipient);
                }
                if (!balancesBefore[order.feeToken][order.owner]) {
                    balancesBefore[order.feeToken][order.owner] =
                        yield this.orderUtil.getERC20Spendable(this.context.tradeDelegate.options.address, order.feeToken, order.owner);
                }
            }
            for (const order of orders) {
                const tokens = [order.tokenS, order.tokenB, order.feeToken];
                for (const token of tokens) {
                    const Token = this.context.ERC20Contract;
                    Token.options.address = order.tokenS;
                    // feeRecipient
                    if (!balancesBefore[token][mining.feeRecipient]) {
                        balancesBefore[token][mining.feeRecipient] =
                            new bignumber_js_1.BigNumber((yield Token.methods.balanceOf(mining.feeRecipient).call()).toString());
                    }
                }
            }
            // Simulate the token transfers of all rings
            const balanceDeltas = {};
            for (const transfer of transferItems) {
                if (!balanceDeltas[transfer.token]) {
                    balanceDeltas[transfer.token] = {};
                }
                if (!balanceDeltas[transfer.token][transfer.from]) {
                    balanceDeltas[transfer.token][transfer.from] = new bignumber_js_1.BigNumber(0);
                }
                if (!balanceDeltas[transfer.token][transfer.to]) {
                    balanceDeltas[transfer.token][transfer.to] = new bignumber_js_1.BigNumber(0);
                }
                balanceDeltas[transfer.token][transfer.from] =
                    balanceDeltas[transfer.token][transfer.from].minus(transfer.amount);
                balanceDeltas[transfer.token][transfer.to] =
                    balanceDeltas[transfer.token][transfer.to].plus(transfer.amount);
            }
            // Calculate the balances after the transaction
            const balancesAfter = {};
            for (const token of Object.keys(balancesBefore)) {
                for (const owner of Object.keys(balancesBefore[token])) {
                    if (!balancesAfter[token]) {
                        balancesAfter[token] = {};
                    }
                    const delta = (balanceDeltas[token] && balanceDeltas[token][owner]) ?
                        balanceDeltas[token][owner] : new bignumber_js_1.BigNumber(0);
                    balancesAfter[token][owner] = balancesBefore[token][owner].plus(delta);
                }
            }
            // Get the fee balances before the transaction
            const feeBalancesBefore = {};
            for (const order of orders) {
                const tokens = [order.tokenS, order.tokenB, order.feeToken];
                for (const token of tokens) {
                    if (!feeBalancesBefore[token]) {
                        feeBalancesBefore[token] = {};
                    }
                    const feeHolderMethods = this.context.feeHolder.methods;
                    // Owner
                    if (!feeBalancesBefore[token][order.owner]) {
                        feeBalancesBefore[token][order.owner] =
                            new bignumber_js_1.BigNumber((yield feeHolderMethods.feeBalances(token, order.owner).call()).toString());
                    }
                    // Wallet
                    if (order.walletAddr && !feeBalancesBefore[token][order.walletAddr]) {
                        feeBalancesBefore[token][order.walletAddr] =
                            new bignumber_js_1.BigNumber((yield feeHolderMethods.feeBalances(token, order.walletAddr).call()).toString());
                    }
                    // FeeRecipient
                    if (!feeBalancesBefore[token][mining.feeRecipient]) {
                        feeBalancesBefore[token][mining.feeRecipient] =
                            new bignumber_js_1.BigNumber((yield feeHolderMethods.feeBalances(token, mining.feeRecipient).call()).toString());
                    }
                    // Burned
                    const feeHolder = this.context.feeHolder.options.address;
                    if (!feeBalancesBefore[token][feeHolder]) {
                        feeBalancesBefore[token][feeHolder] =
                            new bignumber_js_1.BigNumber((yield feeHolderMethods.feeBalances(token, feeHolder).call()).toString());
                    }
                }
            }
            // Calculate the balances after the transaction
            const feeBalancesAfter = {};
            for (const token of Object.keys(feeBalancesBefore)) {
                for (const owner of Object.keys(feeBalancesBefore[token])) {
                    if (!feeBalancesAfter[token]) {
                        feeBalancesAfter[token] = {};
                    }
                    const delta = (feeBalances[token] && feeBalances[token][owner]) ?
                        feeBalances[token][owner] : new bignumber_js_1.BigNumber(0);
                    feeBalancesAfter[token][owner] = feeBalancesBefore[token][owner].plus(delta);
                }
            }
            // Get the filled amounts before
            const filledAmountsBefore = {};
            for (const order of orders) {
                const orderHash = order.hash.toString("hex");
                filledAmountsBefore[orderHash] =
                    new bignumber_js_1.BigNumber((yield this.context.tradeHistory.methods.filled("0x" + orderHash).call()).toString());
            }
            // Filled amounts after
            const filledAmountsAfter = {};
            for (const order of orders) {
                const orderHash = order.hash.toString("hex");
                let filledAmountS = order.filledAmountS ? order.filledAmountS : new bignumber_js_1.BigNumber(0);
                if (!order.valid) {
                    filledAmountS = filledAmountsBefore[orderHash];
                }
                filledAmountsAfter[orderHash] = filledAmountS;
            }
            // Collect the payments
            const payments = {
                rings: [],
            };
            for (const ring of rings) {
                payments.rings.push(ring.payments);
            }
            // Create the report
            const simulatorReport = {
                reverted: false,
                ringMinedEvents,
                invalidRingEvents,
                transferItems,
                feeBalancesBefore,
                feeBalancesAfter,
                filledAmountsBefore,
                filledAmountsAfter,
                balancesBefore,
                balancesAfter,
                payments,
            };
            return simulatorReport;
        });
    }
    validateRings(ringsInfo, report) {
        return __awaiter(this, void 0, void 0, function* () {
            const orders = ringsInfo.orders;
            // Check if we haven't spent more funds than the owner owns
            for (const token of Object.keys(report.balancesAfter)) {
                for (const owner of Object.keys(report.balancesAfter[token])) {
                    assert(report.balancesAfter[token][owner] >= 0, "can't sell more tokens than the owner owns");
                }
            }
            // Check if the spendables were updated correctly
            for (const order of orders) {
                if (order.tokenSpendableS.initialized) {
                    let amountTransferredS = new bignumber_js_1.BigNumber(0);
                    for (const transfer of report.transferItems) {
                        if (transfer.from === order.owner && transfer.token === order.tokenS) {
                            amountTransferredS = amountTransferredS.plus(transfer.amount);
                        }
                    }
                    const amountSpentS = order.tokenSpendableS.initialAmount.minus(order.tokenSpendableS.amount);
                    // amountTransferred could be less than amountSpent because of rebates
                    assert(amountSpentS.gte(amountTransferredS), "amountSpentS >= amountTransferredS");
                }
                if (order.tokenSpendableFee.initialized) {
                    let amountTransferredFee = new bignumber_js_1.BigNumber(0);
                    for (const transfer of report.transferItems) {
                        if (transfer.from === order.owner && transfer.token === order.feeToken) {
                            amountTransferredFee = amountTransferredFee.plus(transfer.amount);
                        }
                    }
                    const amountSpentFee = order.tokenSpendableFee.initialAmount.minus(order.tokenSpendableFee.amount);
                    // amountTransferred could be less than amountSpent because of rebates
                    assert(amountSpentFee.gte(amountTransferredFee), "amountSpentFee >= amountTransferredFee");
                }
            }
            // Check if the allOrNone orders were correctly filled
            for (const order of orders) {
                if (order.allOrNone) {
                    assert(order.filledAmountS.eq(0) || order.filledAmountS.eq(order.amountS), "allOrNone orders should either be completely filled or not at all.");
                }
            }
        });
    }
}
exports.ProtocolSimulator = ProtocolSimulator;
//# sourceMappingURL=protocol_simulator.js.map